#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <avr/pgmspace.h>

#include "Arduino.h"
#include "LiquidCrystal3WI.h"

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set: 
//    DL = 1; 8-bit interface data 
//    N = 0; 1-line display 
//    F = 0; 5x8 dot character font 
// 3. Display on/off control: 
//    D = 0; Display off 
//    C = 0; Cursor off 
//    B = 0; Blinking off 
// 4. Entry mode set: 
//    I/D = 1; Increment by 1 
//    S = 0; No shift 
//
// Note, however, that resetting the Arduino doesn't reset the LCD, so we
// can't assume that its in that state when a sketch starts (and the
// LiquidCrystal constructor is called).

PROGMEM const char toPage1_1[] = {	0x41, 0xA0, 0x42, 0xA1, 0xE0, 0x45, 0xA3, 0xA4, 0xA5, 0xA6, /* А Б В Г Д Е Ж З И Й */
									0x4B, 0xA7, 0x4D, 0x48, 0x4F, 0xA8, 0x50, 0x43, 0x54, 0xA9, /* К Л М Н О П Р С Т У */
									0xAA, 0x58, 0xE1, 0xAB, 0xAC, 0xE2, 0xAD, 0xAE, 0x08, 0xAF, /* Ф Х Ц Ч Ш Щ Ъ Ы Ь Э */
									0xB0, 0xB1, 0x61, 0xB2, 0xB3, 0xB4, 0xE3, 0x65, 0xB6, 0xB7, /* Ю Я а б в г д е ж з */
									0xB8, 0xB9, 0xBA, 0xBB, 0xBC, 0xBD, 0x6F, 0xBE				/* и й к л м н о п */
									};

PROGMEM const char toPage1_2[] = {	0x70, 0x63, 0xBF, 0x79, 0xE4, 0x78, 0xE5, 0xC0, 0xC1, 0xE6, /* р с т у ф х ц ч ш щ */
									0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7							/* ъ ы ь э ю я */
									};

LiquidCrystal3WI::LiquidCrystal3WI(uint8_t dataPin, uint8_t clockPin, uint8_t latchPin)
{
  init(dataPin, clockPin, latchPin, true);
}

LiquidCrystal3WI::LiquidCrystal3WI(uint8_t dataPin, uint8_t clockPin, uint8_t latchPin, bool usePage2)
{
  init(dataPin, clockPin, latchPin, usePage2);
}

void LiquidCrystal3WI::init(uint8_t dataPin, uint8_t clockPin, uint8_t latchPin, bool usePage2)
{
	_dataPin = dataPin;
	_clockPin = clockPin;
	_latchPin = latchPin;

	_registerData = 0;

	_backlignt = true;
	_usePage2 = usePage2;

	pinMode(this->_dataPin, OUTPUT);
	digitalWrite(this->_dataPin, LOW);
	pinMode(this->_clockPin, OUTPUT);
	digitalWrite(this->_clockPin, LOW);
	pinMode(this->_latchPin, OUTPUT);
	digitalWrite(this->_latchPin, LOW);

	_displayfunction = LCD_4BITMODE | LCD_1LINE | LCD_5x8DOTS;

	if (_usePage2) {
		_displayfunction |= 2;
	}
  
	begin(16, 1);
}

void LiquidCrystal3WI::begin(uint8_t cols, uint8_t lines, uint8_t dotsize) {
	if (lines > 1) {
		_displayfunction |= LCD_2LINE;
	}
	_numlines = lines;

	setRowOffsets(0x00, 0x40, 0x00 + cols, 0x40 + cols);

	// for some 1 line displays you can select a 10 pixel high font
	if ((dotsize != LCD_5x8DOTS) && (lines == 1)) {
		_displayfunction |= LCD_5x10DOTS;
	}

	// SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
	// according to datasheet, we need at least 40ms after power rises above 2.7V
	// before sending commands. Arduino can turn on way before 4.5V so we'll wait 50
	delayMicroseconds(50000);
	// Now we pull both RS and R/W low to begin commands
	setupDisplayPins(LOW, LOW, LOW, 0, true);

	// this is according to the hitachi HD44780 datasheet
	// figure 24, pg 46

	// we start in 8bit mode, try to set 4 bit mode
	write4bits(0x03);
	delayMicroseconds(4500); // wait min 4.1ms

	// second try
	write4bits(0x03);
	delayMicroseconds(4500); // wait min 4.1ms

	// third go!
	write4bits(0x03);
	delayMicroseconds(150);

	// finally, set to 4-bit interface
	write4bits(0x02);

	// finally, set # lines, font size, etc.
	command(LCD_FUNCTIONSET | _displayfunction);

	// turn the display on with no cursor or blinking default
	_displaycontrol = LCD_DISPLAYON | LCD_CURSOROFF | LCD_BLINKOFF;
	display();

	// clear it off
	clear();

	// Initialize to default text direction (for romance languages)
	_displaymode = LCD_ENTRYLEFT | LCD_ENTRYSHIFTDECREMENT;
	// set the entry mode
	command(LCD_ENTRYMODESET | _displaymode);

}

void LiquidCrystal3WI::setBacklight(bool on) {
	_backlignt = on;
	setupDisplayPins();
}

void LiquidCrystal3WI::printUTF8(const char *utf8String) {
	const unsigned char *from = (unsigned char *) utf8String;
	for (size_t i = 0; from[i] != 0; i++) {
		if ((from[i] == 0xD0) || (from[i] == 0xD1)) {
			uint16_t ch = from[i++] * 0x100;
			ch += from[i];
			printUTF8(ch & 0xFFFF);
		} else {
			print(utf8String[i]);
		}
	}
}

void LiquidCrystal3WI::printUTF8(uint16_t ch) {
	if (_usePage2) {
		if ((ch >= 0xD090) && (ch <= 0xD0BF)) { // small letters 'а' .. 'п'
			ch &= 0xFF;
			print((char) (ch + 0x30));
		} else if ((ch >= 0xD180) && (ch <= 0xD18F)) { // small letters 'р' .. 'я'
			ch &= 0xFF;
			print((char) (ch + 0x70));
		} else if (ch == 0xD081) { // letter 'Ё'
			print((char) 0xA8);
		} else if (ch == 0xD191) { // letter 'ё'
			print((char) 0xB8);
		} else {
			print((char) ch);
		}
	} else {
		if ((ch >= 0xD090) && (ch <= 0xD0BF)) { // small letters 'а' .. 'п'
			ch &= 0xFF;
			print((char) pgm_read_byte(&toPage1_1[ch - 0x90]));
		} else if ((ch >= 0xD180) && (ch <= 0xD18F)) { // small letters 'р' .. 'я'
			ch &= 0xFF;
			print((char) pgm_read_byte(&toPage1_2[ch - 0x80]));
		} else if (ch == 0xD081) { // letter 'Ё'
			print((char) 0xA2);
		} else if (ch == 0xD191) { // letter 'ё'
			print((char) 0xB5);
		} else {
			print((char) ch);
		}
	}
}

/*
   in some 16x4 LCD when line 3 and 4 are not placed correctly you may try:
     setRowOffsets(0x00, 0x40, 0x14, 0x54)
   or
     setRowOffsets(0x00, 0x40, 0x10, 0x50)
 */
void LiquidCrystal3WI::setRowOffsets(int row0, int row1, int row2, int row3)
{
  _row_offsets[0] = row0;
  _row_offsets[1] = row1;
  _row_offsets[2] = row2;
  _row_offsets[3] = row3;
}

/********** high level commands, for the user! */
void LiquidCrystal3WI::clear()
{
  command(LCD_CLEARDISPLAY);  // clear display, set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void LiquidCrystal3WI::home()
{
  command(LCD_RETURNHOME);  // set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void LiquidCrystal3WI::setCursor(uint8_t col, uint8_t row)
{
  const size_t max_lines = sizeof(_row_offsets) / sizeof(*_row_offsets);
  if ( row >= max_lines ) {
    row = max_lines - 1;    // we count rows starting w/0
  }
  if ( row >= _numlines ) {
    row = _numlines - 1;    // we count rows starting w/0
  }
  
  command(LCD_SETDDRAMADDR | (col + _row_offsets[row]));
}

// Turn the display on/off (quickly)
void LiquidCrystal3WI::noDisplay() {
  _displaycontrol &= ~LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal3WI::display() {
  _displaycontrol |= LCD_DISPLAYON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turns the underline cursor on/off
void LiquidCrystal3WI::noCursor() {
  _displaycontrol &= ~LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal3WI::cursor() {
  _displaycontrol |= LCD_CURSORON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// Turn on and off the blinking cursor
void LiquidCrystal3WI::noBlink() {
  _displaycontrol &= ~LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}
void LiquidCrystal3WI::blink() {
  _displaycontrol |= LCD_BLINKON;
  command(LCD_DISPLAYCONTROL | _displaycontrol);
}

// These commands scroll the display without changing the RAM
void LiquidCrystal3WI::scrollDisplayLeft(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVELEFT);
}
void LiquidCrystal3WI::scrollDisplayRight(void) {
  command(LCD_CURSORSHIFT | LCD_DISPLAYMOVE | LCD_MOVERIGHT);
}

// This is for text that flows Left to Right
void LiquidCrystal3WI::leftToRight(void) {
  _displaymode |= LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void LiquidCrystal3WI::rightToLeft(void) {
  _displaymode &= ~LCD_ENTRYLEFT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'right justify' text from the cursor
void LiquidCrystal3WI::autoscroll(void) {
  _displaymode |= LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// This will 'left justify' text from the cursor
void LiquidCrystal3WI::noAutoscroll(void) {
  _displaymode &= ~LCD_ENTRYSHIFTINCREMENT;
  command(LCD_ENTRYMODESET | _displaymode);
}

// Allows us to fill the first 8 CGRAM locations
// with custom characters
void LiquidCrystal3WI::createChar(uint8_t location, uint8_t charmap[]) {
  location &= 0x7; // we only have 8 locations 0-7
  command(LCD_SETCGRAMADDR | (location << 3));
  for (int i=0; i<8; i++) {
    write(charmap[i]);
  }
}

/*********** mid level commands, for sending data/cmds */

inline void LiquidCrystal3WI::command(uint8_t value) {
  send(value, LOW);
}

inline size_t LiquidCrystal3WI::write(uint8_t value) {
	send(value, HIGH);
	return 1; // assume sucess
}

/************ low level data pushing commands **********/

// write either command or data, with automatic 4/8-bit selection
void LiquidCrystal3WI::send(uint8_t value, uint8_t mode) {
	//	digitalWrite(_rs_pin, mode);
	setupDisplayPins(-1, mode, -1, -1, false);

	// if there is a RW pin indicated, set it low to Write
	//	if (_rw_pin != 255) {
	//		digitalWrite(_rw_pin, LOW);
	//	}
	setupDisplayPins(-1, -1, LOW, -1, false);

	write4bits(value >> 4);
	write4bits(value);
}

void LiquidCrystal3WI::pulseEnable(void) {
	//	digitalWrite(_enable_pin, LOW);
	setupDisplayPins(LOW, -1, -1, -1, false);
	delayMicroseconds(1);
	//	digitalWrite(_enable_pin, HIGH);
	setupDisplayPins(HIGH, -1, -1, -1, false);
	delayMicroseconds(1);    // enable pulse must be >450ns
	//	digitalWrite(_enable_pin, LOW);
	setupDisplayPins(LOW, -1, -1, -1, false);
	delayMicroseconds(100);   // commands need > 37us to settle
}

void LiquidCrystal3WI::write4bits(uint8_t value) {
	setupDisplayPins(-1, -1, -1, value, true);
	pulseEnable();
}

void LiquidCrystal3WI::setupDisplayPins(uint8_t enablePin, uint8_t resetPin, uint8_t rwPin, uint8_t data, bool setData) {
	setupRegisterData(enablePin, resetPin, rwPin, data, setData);
	setupDisplayPins();
}

void LiquidCrystal3WI::setupRegisterData(uint8_t enablePin, uint8_t resetPin, uint8_t rwPin, uint8_t data, bool setData) {
	if (enablePin == 1) {
		_registerData |= 0b10000000;
	} else if (enablePin == 0) {
		_registerData &= 0b01111111;
	}
	if (rwPin == 1) {
		_registerData |= 0b01000000;
	} else if (rwPin == 0) {
		_registerData &= 0b10111111;
	}
	if (resetPin == 1) {
		_registerData |= 0b00100000;
	} else if (resetPin == 0) {
		_registerData &= 0b11011111;
	}
	if (setData) {
		data &= 0x0F;
		_registerData &= 0b11100000;
		_registerData |= data * 2;
	}
}

void LiquidCrystal3WI::setupDisplayPins() {
	if (_backlignt) {
		_registerData |= 1;
	} else {
		_registerData &= 0b11111110;
	}
	for (uint8_t i = 0b10000000; i > 0; i /= 2) {
		uint8_t b = i & _registerData ? HIGH : LOW;
		SEND_BIT(this->_dataPin, this->_clockPin, b);
	}
	digitalWrite(_latchPin, HIGH);
	digitalWrite(_latchPin, LOW);
}
